#include <math.h>
#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"
#include "ggi_widget_highlevel.h"

#define MAXNUMLEN 16

struct checkboxdata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
	int    is_inside,will_activate;
	int	currvalue;
};

static int checkbox_getsize(ggi_widget_t self) {

	ggi_widget_t curchild;
	
	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;
	
	if ((curchild=self->getchild(self,0))) {
		self->min=curchild->min;
		self->opt=curchild->opt;
	}
	return 0;
}

static int ggiWidgetCheckboxRedraw(ggi_widget_t self,int flush) {

	struct checkboxdata *mine=self->priv;
	int rc;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);
	rc=ggiWidgetRedrawChildren(self);
	if (flush)
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	return rc;
}

static int ggiWidgetCheckboxVischanged(ggi_widget_t self) {

	struct checkboxdata *mine=self->priv;
	ggi_widget_t curchild;
	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,1,1);

	if ((curchild=self->getchild(self,0))) {
		curchild->vis=self->vis;
		curchild->pos.x =self->pos.x+mine->drawpos.x;
		curchild->pos.y =self->pos.y+mine->drawpos.y;
		curchild->size.x=mine->drawsize.x;
		curchild->size.y=mine->drawsize.y;
		curchild->vischanged(curchild);
	}
	return 0;
}

static int ggiWidgetCheckboxChildchanged(ggi_widget_t self,ggi_widget_t child) {

	checkbox_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

static void cb_new_value(ggi_widget_t self) {

	struct checkboxdata *mine=self->priv;

	mine->currvalue=!!mine->currvalue;
	if (self->statevar)
		*((int *)self->statevar)=mine->currvalue;
	self->redraw(self,1);
}

int ggiWidgetCheckboxEvent(ggi_widget_t self,gii_event *ev,ggi_widget_inputstatus_t status) {

	struct checkboxdata *mine=self->priv;
	int is_inside;

	is_inside=(status->mouse.x>=self->pos.x+mine->drawpos.x &&
		   status->mouse.x< self->pos.x+mine->drawpos.x+mine->drawsize.x &&
		   status->mouse.y>=self->pos.y+mine->drawpos.y &&
		   status->mouse.y< self->pos.y+mine->drawpos.y+mine->drawsize.y );

	if (is_inside&&!mine->is_inside) {
		if (self->callback) {
			self->callback(GWT_CB_ENTER,self);
		}
	} else if (!is_inside&&mine->is_inside) {
		if (self->callback) {
			self->callback(GWT_CB_LEAVE,self);
		}
	};
	mine->is_inside=is_inside;

	switch(ev->any.type) {
		case evKeyPress:
		case evKeyRepeat:
			if (ev->key.sym==self->hotkey) {
				mine->will_activate=1;
				toggle:
				mine->currvalue=!mine->currvalue;
				cb_new_value(self);
				if (self->callback) {
					self->callback(GWT_CB_STATECHANGE,self);
				}
				return 1;
			}
			if (self->has_focus&&ev->key.sym==' ') {
				goto toggle;
			}
			break;
		case evKeyRelease:
			mine->will_activate=0;
			break;
		case evPtrButtonPress:
			if (is_inside) {
				ggiWidgetUnfocus(self);
				self->has_focus=1;
				if (ev->pbutton.button==GII_PBUTTON_PRIMARY) {
					mine->will_activate=1;
				}
				self->redraw(self,1);
			}
			break;
		case evPtrButtonRelease:
			if (is_inside) {
				if (ev->pbutton.button==GII_PBUTTON_PRIMARY&&
				    mine->will_activate) {
					mine->will_activate=0;
					self->redraw(self,1);
					goto toggle;
				}
			}
			if (mine->will_activate) {
				mine->will_activate=0;
				self->redraw(self,1);
			}
			break;
	}
	return 0;
}


ggi_widget_t ggiWidgetCreateCheckbox(va_list ap) {

	ggi_widget_t newid,cline,card,checked,unchecked,parms;
	struct checkboxdata *newcheckbox;

	static struct ggiWidgetBitmap checkedbm={
		{8,8},"\xab\x03\x87\x0f\xdc\xf9\xf0\x75"
	};
	static struct ggiWidgetBitmap uncheckedbm={
		{8,8},"\xaa\x01\x80\x01\x80\x01\x80\x55"
	};

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newcheckbox=malloc(sizeof(*newcheckbox)))) {
		free(newid);
		return NULL;
	}
	newcheckbox->is_inside=0;
	newcheckbox->will_activate=0;
	newcheckbox->currvalue=0;
	newid->priv=newcheckbox;

	newid->max.x=newid->max.y=-1;

	newid->redraw    =ggiWidgetCheckboxRedraw;
	newid->vischanged=ggiWidgetCheckboxVischanged;
	newid->childchanged=ggiWidgetCheckboxChildchanged;
	newid->process_event=ggiWidgetCheckboxEvent;

	newid->getchild   =ggiWidgetSingleChildGet;
	newid->linkchild  =ggiWidgetSingleChildLink;
	newid->unlinkchild=ggiWidgetSingleChildUnlink;

	unchecked=ggiWidgetCreateWidget(GWT_BITMAP,&uncheckedbm);
	checked  =ggiWidgetCreateWidget(GWT_BITMAP,&checkedbm);

	card     =ggiWidgetCreateWidget(GWT_CONTAINER_CARD,unchecked,checked,NULL);
	card->statevar=&newcheckbox->currvalue;

	cline    =ggiWidgetCreateWidget(GWT_CONTAINER_LINE,2,1,card,NULL);

	while( (parms=va_arg(ap,ggi_widget_t)) ) {
		cline->linkchild(cline,LAST_CHILD,parms);
	}
	newid->linkchild(newid,0,cline);
	checkbox_getsize(newid);

	return newid;
}
