
#include <stdarg.h>

#include "ggi_widget.h"
#include "ggi_widget_internal.h"

#define DOUBLECLICK_MS   300
#define DOUBLECLICK_MOVE   5

static int process_event_dummy(ggi_widget_t self,gii_event *ev,ggi_widget_inputstatus_t stat) {

	int x,rc;
	ggi_widget_t child;
	
	rc=0;
	for(x=0;(child=self->getchild(self,x));x++) {
		rc|=child->process_event(child,ev,stat);
	}
	return rc;
}

static ggi_widget_t getchild_dummy(ggi_widget_t unused,int unused2) {
	return NULL;
}

static int linkchild_dummy(ggi_widget_t unused,int unused2, ggi_widget_t unused3,...) {
	return 0;
}

static int vischanged_dummy(ggi_widget_t unused) {
	return 0;
}

static int childchanged_dummy(ggi_widget_t unused,ggi_widget_t unused2) {
	return 0;
}

static int destroy_dummy(ggi_widget_t self) {

	int rc;
	
	if (!self) return 1;

	rc=ggiWidgetDestroyChildren(self);

	if (self->priv) free(self->priv);
	free(self);
	return rc;
}

ggi_widget_t ggiWidgetAllocate(void) {

	ggi_widget_t wa;
	
	wa=malloc(sizeof(*wa));
	if (!wa) return NULL;
	
	wa->parent=NULL;
	wa->colorset=NULL;

	wa->min.x=wa->min.y=0;
	wa->opt.x=wa->opt.y=0;
	wa->max.x=wa->max.y=-1;

	wa->gravity=GWT_GRAV_CENTER;
	wa->entered=0;
	wa->has_focus=0;
	wa->greyed_out=0;

	wa->vis=NULL;
	wa->pos.x=wa->pos.y=0;
	wa->size.x=wa->size.y=0;

	wa->hotkey=GIIK_VOID;
	wa->statevar=NULL;
	wa->callback=NULL;
	wa->callbackpriv=NULL;
	
	wa->destroy      =destroy_dummy;
	wa->redraw       =NULL;
	wa->vischanged   =vischanged_dummy;
	wa->childchanged =childchanged_dummy;

	wa->process_event=process_event_dummy;
	
	wa->children   =NULL;
	wa->getchild   =getchild_dummy;
	wa->linkchild  =linkchild_dummy;
	wa->unlinkchild=getchild_dummy;
	
	wa->priv=NULL;

	return wa;
}

ggi_widget_t ggiWidgetCreateWidget(ggiWidgetType type,...) {

	va_list ap;
	ggi_widget_t rc;
	
	va_start(ap,type);
	
	switch(type) {
		case GWT_LABEL:
			rc=ggiWidgetCreateLabel(ap);
			break;
		case GWT_BITMAP:
			rc=ggiWidgetCreateBitmap(ap);
			break;
		case GWT_FRAME:
			rc=ggiWidgetCreateFrame(ap);
			break;
		case GWT_BUTTON:
			rc=ggiWidgetCreateButton(ap);
			break;
		case GWT_CHECKBOX:
			rc=ggiWidgetCreateCheckbox(ap);
			break;
		case GWT_OPTIONCYCLER:
			rc=ggiWidgetCreateOptioncycler(ap);
			break;
		case GWT_TEXT:
			rc=ggiWidgetCreateText(ap);
			break;
		case GWT_ODOMETER:
			rc=ggiWidgetCreateOdometer(ap);
			break;
		case GWT_DIAL:
			rc=ggiWidgetCreateDial(ap);
			break;
		case GWT_CONTAINER_LINE:
			rc=ggiWidgetCreateContainerLine(ap);
			break;
		case GWT_CONTAINER_STACK:
			rc=ggiWidgetCreateContainerStack(ap);
			break;
		case GWT_CONTAINER_GRID:
			rc=ggiWidgetCreateContainerGrid(ap);
			break;
		case GWT_CONTAINER_CARD:
			rc=ggiWidgetCreateContainerCard(ap);
			break;
		default:
			rc=NULL;
			break;
	}
	
	va_end(ap);

	return rc;
}


static int timediff_ms(struct timeval *a,struct timeval *b) {

	int rc;
	
	rc =(a->tv_sec -b->tv_sec )*1000;
	rc+=(a->tv_usec-b->tv_usec)/1000;
	return rc;
}

static ggi_widget_inputstatus_t inputroot=NULL;

int ggiWidgetProcessEvent(ggi_visual_t vis,gii_event *ev) {

	ggi_widget_inputstatus_t curr;
	ggi_widget_childlist_t cchild;	
	int rc;
	
	for(curr=inputroot;curr;curr=curr->next) {
		if (curr->vis==vis) goto goon;
	}
	return 0;

	goon:
	switch(ev->any.type) {
		case evKeyPress:
		case evKeyRepeat:
		case evKeyRelease:
			curr->key.modifiers=ev->key.modifiers;
			break;
		case evPtrAbsolute:
			curr->mouse.x    =ev->pmove.x;
			curr->mouse.y    =ev->pmove.y;
			curr->mouse.z    =ev->pmove.z;
			curr->mouse.wheel=ev->pmove.wheel;
			break;
		case evPtrButtonPress:
			if (ev->pbutton.button<31) {
				curr->mouse.buttonmask|=(1<<ev->pbutton.button);
			}
			if (DOUBLECLICK_MS>timediff_ms(&ev->pbutton.time,&curr->mouse.lastclicktime)) {
				int dx,dy;
				dx=curr->mouse.x-curr->mouse.lcx;
				dy=curr->mouse.y-curr->mouse.lcy;
				if (dx<0) dx=-dx;
				if (dy<0) dy=-dy;
				if (dx<DOUBLECLICK_MOVE&&dy<DOUBLECLICK_MOVE) {
					curr->mouse.clickcount++;
				} else {
					curr->mouse.clickcount=1;
				}
			}
			curr->mouse.lastclicktime=ev->pbutton.time;
			curr->mouse.lcx=curr->mouse.x;
			curr->mouse.lcy=curr->mouse.y;
			break;
		case evPtrButtonRelease:
			if (ev->pbutton.button<31) {
				curr->mouse.buttonmask&=~(1<<ev->pbutton.button);
			}
			break;
	}
	for(cchild=curr->children;cchild;cchild=cchild->next) {
		rc|=cchild->child->process_event(cchild->child,ev,curr);
	}
	return rc;
}

int ggiWidgetAttach(ggi_widget_t widget, ggi_visual_t vis,int x,int y,int w,int h) {

	ggi_widget_inputstatus_t curr;
	ggi_widget_childlist_t cchild;	

	if (NULL==(cchild=malloc(sizeof(*cchild)))) {
		return 1;
	}
	cchild->child=widget;

	for(curr=inputroot;curr;curr=curr->next) {
		if (curr->vis==vis) goto goon;
	}
	/* Nothing found. Create new one. */
	if (NULL==(curr=malloc(sizeof(*curr)))) {
		free(cchild);
		return 1;
	}
	curr->vis=vis;
	curr->children=NULL;
	curr->mouse.x=
	curr->mouse.y=
	curr->mouse.z=
	curr->mouse.wheel=
	curr->mouse.lcx=
	curr->mouse.lcy=
	curr->mouse.clickcount=0;
	curr->mouse.lastclicktime.tv_sec=0;
	curr->mouse.lastclicktime.tv_usec=0;
	curr->next=inputroot;
	inputroot=curr;

	goon:
	cchild->next=curr->children;
	curr->children=cchild;

	widget->vis=vis;
	widget->pos.x=x;
	widget->pos.y=y;
	widget->size.x=w;
	widget->size.y=h;
	widget->vischanged(widget);

	return 0;
}

int ggiWidgetDetach(ggi_widget_t widget, ggi_visual_t vis) {

	ggi_widget_inputstatus_t curr,*curptr;
	ggi_widget_childlist_t cchild,*chrptr;

	for(curr=*(curptr=&inputroot);curr;curr=*(curptr=&curr->next)) {
		if (curr->vis==vis) goto goon;
	}
	return 1;
	goon:

	for(cchild=*(chrptr=&curr->children);cchild;cchild=*(chrptr=&cchild->next)) {
		if (cchild->child==widget) {
			*chrptr=cchild->next;
			free(cchild);
			if (curr->children==NULL) {
				*curptr=curr->next;
				free(curr);
			}
			widget->vis=NULL;
			widget->pos.x=0;
			widget->pos.y=0;
			widget->size.x=0;
			widget->size.y=0;
			widget->vischanged(widget);
			return 0;
		}
	}
	return 1;
}

ggi_widget_color_set *ggiWidgetGetColorset(ggi_widget_t widget) {

	for(;widget;widget=widget->parent) {
		if (widget->colorset) 
			return widget->colorset;
	}
	return &ggi_widget_default_colorset;
}

ggi_widget_t findfocus(ggi_widget_t widget) {
	int x;
	ggi_widget_t currwid;
	if (widget->has_focus)
		return widget;
	for(x=0;(currwid=widget->getchild(widget,x));x++) {
		currwid=findfocus(currwid);
		if (currwid) return currwid;
	}
	return NULL;
}

ggi_widget_t ggiWidgetGetCurrentFocus(ggi_widget_t widget) {

	if (widget->has_focus) 
		return widget;
	while(widget->parent) widget=widget->parent;
	return findfocus(widget);
}

ggi_widget_t ggiWidgetUnfocus(ggi_widget_t widget) {

	ggi_widget_t oldfocus;
	
	oldfocus=ggiWidgetGetCurrentFocus(widget);
	
	if (oldfocus) {
		oldfocus->has_focus=0;
		oldfocus->redraw(oldfocus,1);
	}
	return oldfocus;
}
